% dfAlign - Affine image alignment using distribution fields
%   [fit, xySigma, fSigma,lik] = dfAlign(img, tmplt, p_init, NUM_ITERATIONS,downsampleScale,...
%    USE_SIC,numBins,sigmaXYChoices,sigmaFChoices,direction,additive)
%   Align the template image tmplt to an example image img using an
%   affine warp initialized using p_init.
%
%   This assumes grayscale images and rectangular templates.
%
%    Inputs:
%        img: The image being aligned to tmplt
%        tmplt: The region that img is being aligned to
%        p_init: The initial warp to apply to img
%        NUM_ITERATIONS: Maximum number of iterations
%        downsampleScale: How much to downsample the distribution
%        fields in the spatial dimensions
%        USE_SIC: Whether to perform a simultaneous search over bias and
%        gain parameters
%        numBins: Number of bins used to construct the distribution
%        fields
%        sigmaXYChoices: Set of xy-sigmas to search over when
%        constructing the distribution fields
%        sigmaFChoices: Set of feature space sigmas to search over when
%        constructing the distribution fields.
%        direction: whether to perform forward (0),
%        inverse (1), or inverse-forward (2);
%        additive: Whether to perform additive updates (1) or
%        compositional updates (0)
%
%    Outputs:
%        fit: Struct containing the intermediate and final transformations
%        xySigma: Final xySigma that maximized the likelihood of warped input
%        image under DF of template
%        fSigma: Final fSigma that maximized the likelihood of warped input
%        image under DF of template
%        lik: Final likelihood computed using parameters that maximized the
%        likelihood
%

function [fit, xySigma, fSigma,lik] = dfAlign(img, tmplt, p_init, NUM_ITERATIONS,downsampleScale,...
    USE_SIC,numBins,sigmaXYChoices,sigmaFChoices,direction,additive)

%Note: if numBins ==1, then use the Lucas-Kanade method.
if(~exist('additive','var')) additive = false; end
if(~exist('direction','var')) direction = 2; end
if(additive && (numBins~=1 || direction~=0 || USE_SIC))
    error('Additive updates can only be used with Lucas Kanade forward algorithm');
end

fit = struct([]);
height = size(tmplt,1); width = size(tmplt,2);

T = [p_init' [0;0;1]];

F_KERNEL_SIZE = 20; %always use truncated kernel size of 20 for feature space dimension
heightNew = floor(height/downsampleScale);
widthNew = floor(width/downsampleScale);

%Convert p_init to downsampled coord system
pntsOrig = [1 1 1;1 height 1;width 1 1;width height 1];
pntsNew = pntsOrig* T;
pntsNew = pntsNew./repmat(pntsNew(:,3),[1 3]);
downSamplePntsOrig = pntsOrig/downsampleScale;
downSamplePntsOrig(:,3)=1;
downSampledPntsNew = [ pntsNew(:,1:2)/downsampleScale ones(4,1)];
T = pinv(downSamplePntsOrig)*downSampledPntsNew;


if(numBins ~= 1)
    %precompute distribution fields for template
    templateDistributionFields = zeros(heightNew,widthNew,numBins,size(sigmaXYChoices,2),size(sigmaFChoices,2));
    templateDistributionFields2 = zeros(height,width,numBins,size(sigmaXYChoices,2),size(sigmaFChoices,2));
    
    %%A1 is for bias and A2 for gain
    templateDistributionFieldsA1 = zeros(heightNew,widthNew,numBins,size(sigmaXYChoices,2),size(sigmaFChoices,2));
    templateDistributionFieldsA2 = zeros(heightNew,widthNew,numBins,size(sigmaXYChoices,2),size(sigmaFChoices,2));
    for i=1:size(sigmaXYChoices,2)
        tempDF = explodePadUniformXY(tmplt,sigmaXYChoices(i),max(4*sigmaXYChoices(i)+1,3),...
            sigmaXYChoices(i),max(3,4*sigmaXYChoices(i)+1),numBins);
        
        
        if(USE_SIC)
            tempDFA1 = explodePadUniformXY(min(255,tmplt+1),sigmaXYChoices(i),max(4*sigmaXYChoices(i)+1,3),...
                sigmaXYChoices(i),max(3,4*sigmaXYChoices(i)+1),numBins);
            tempDFA2 = explodePadUniformXY(min(255,tmplt*1.01),sigmaXYChoices(i),max(4*sigmaXYChoices(i)+1,3),...
                sigmaXYChoices(i),max(3,4*sigmaXYChoices(i)+1),numBins);
        end
        
        for j=1:size(sigmaFChoices,2)
            filter3=reshape(makeGaussianKernel(F_KERNEL_SIZE, ...
                sigmaFChoices(j)/(256/numBins)),[1 1 F_KERNEL_SIZE]);
            tempDF2 =convn(tempDF, filter3, 'same');
            templateDistributionFields2(:,:,:,i,j) = tempDF2;
            templateDistributionFields(:,:,:,i,j) = tempDF2(downsampleScale:downsampleScale:end,downsampleScale:downsampleScale:end,:);
            if(USE_SIC)
                tempDF2 =convn(tempDFA2, filter3, 'same');
                templateDistributionFieldsA2(:,:,:,i,j) = tempDF2(downsampleScale:downsampleScale:end,downsampleScale:downsampleScale:end,:);
                tempDF2 =convn(tempDFA1, filter3, 'same');
                templateDistributionFieldsA1(:,:,:,i,j) = tempDF2(downsampleScale:downsampleScale:end,downsampleScale:downsampleScale:end,:);
            end
        end
    end
else %use regular Lucas-Kanade method
    template = tmplt;
    templateA1 = tmplt+1;
    templateA2 = tmplt*1.01;
end


if(USE_SIC) degFreedom = 6+2;
else degFreedom = 6; end

if(USE_SIC) dWdP = zeros(2,degFreedom-2,heightNew,widthNew);
else dWdP = zeros(2,degFreedom,heightNew,widthNew); end

dWdPAll = zeros(2*numBins*heightNew*widthNew,6);
for j=1:widthNew
    for i=1:heightNew
        dWdP(:,:,i,j)=[j 0 i 0 1 0;
            0 j 0 i 0 1];
        dWdPAll((i-1)*numBins*2+heightNew*numBins*2*(j-1)+1:...
            (i-1)*numBins*2+heightNew*numBins*2*(j-1)+numBins*2,:) = ...
            repmat([j 0 i 0 1 0;0 j 0 i 0 1],[numBins 1]);
    end
end

dWdPReshaped = zeros(heightNew*widthNew*numBins,2,6);
for i=1:6
    dWdPReshaped(:,:,i) = reshape(dWdPAll(:,i),2,[])';
end

%Precompute Steepest Descent Images and Hessian
featureSpaceDim = numBins;
if(numBins~=1)
    steepestDescentImages = zeros(heightNew*widthNew*featureSpaceDim,degFreedom,...
        size(sigmaXYChoices,2),size(sigmaFChoices,2));
    Hessians = zeros(degFreedom,degFreedom,size(sigmaXYChoices,2),size(sigmaFChoices,2));
    InverseHessians = zeros(degFreedom,degFreedom,size(sigmaXYChoices,2),size(sigmaFChoices,2));
    
    for m=1:size(sigmaXYChoices,2)
        for n=1:size(sigmaFChoices,2)
            [Ex,Ey] = gradient(templateDistributionFields(:,:,:,m,n));
            if(USE_SIC)
                EA1 = templateDistributionFieldsA1(:,:,:,m,n) - templateDistributionFields(:,:,:,m,n);
                EA2 = (templateDistributionFieldsA2(:,:,:,m,n) - templateDistributionFields(:,:,:,m,n))/.01;
            end
            ExReshaped = shiftdim(Ex,2); EyReshaped = shiftdim(Ey,2);
            ExEy = [ExReshaped(:) EyReshaped(:)];
            for i=1:6
                steepestDescentImages(:,i,m,n) = sum(ExEy.*dWdPReshaped(:,:,i),2);
            end
            if(USE_SIC)
                EA1Reshaped = shiftdim(EA1,2); EA2Reshaped = shiftdim(EA2,2);
                steepestDescentImages(:,7,m,n) = EA1Reshaped(:);
                steepestDescentImages(:,8,m,n) = EA2Reshaped(:);
            end
            Hessians(:,:,m,n) = steepestDescentImages(:,:,m,n)'*steepestDescentImages(:,:,m,n);
        end
    end
    
    %Precompute Inverses of Hessians
    for m=1:size(sigmaXYChoices,2)
        for n=1:size(sigmaFChoices,2)
            InverseHessians(:,:,m,n) = Hessians(:,:,m,n)^(-1);
        end
    end
else %use LK method
    [Ex,Ey] = gradient(template);
    if(USE_SIC)
        EA1 = templateA1 - template;
        EA2 = (templateA2- template)/.01;
    end
    ExEy = [Ex(:) Ey(:)];
    steepestDescentImage = zeros(heightNew*widthNew*featureSpaceDim,degFreedom);
    for i=1:6
        steepestDescentImage(:,i) = sum(ExEy.*dWdPReshaped(:,:,i),2);
    end
    if(USE_SIC)
        steepestDescentImage(:,7) = EA1(:);
        steepestDescentImage(:,8) = EA2(:);
    end
    Hessian = steepestDescentImage'*steepestDescentImage;
    InverseHessian = Hessian^(-1);
end

additiveNoise = 0; gain = 0;
iteration =1;
while(true)
    %Compute warp in non-downsampled image
    downSampledPntsOrig = [1 1 1;1 heightNew 1;widthNew 1 1;widthNew heightNew 1];
    downSampledPntsNew = downSampledPntsOrig* T;
    downSampledPntsNew = downSampledPntsNew./repmat(downSampledPntsNew(:,3),[1 3]);
    pntsOrig = [downsampleScale* downSampledPntsOrig(:,1:2) ones(4,1)];
    pntsNew = [downsampleScale* downSampledPntsNew(:,1:2) ones(4,1)];
    T_temp = pinv(pntsOrig)*pntsNew;
    fit(iteration).warp_p = T_temp'; %save transpose to agree with Lucas-Kanade code convention
    
    currentWarpedImage = (imtransform2(img,T_temp)*(gain+1))+additiveNoise;
    currentWarpedImage = currentWarpedImage(1:height,1:width);
    
    currentWarpedImage(currentWarpedImage>255)=255;
    currentWarpedImage(currentWarpedImage<0)=0;
    
    if(numBins~=1)
        [bestXYChoice,bestFChoice] = findBestSigma(templateDistributionFields2,currentWarpedImage,...
            width,height,sigmaXYChoices,sigmaFChoices,numBins);
        fit(iteration).xySigma = sigmaXYChoices(bestXYChoice);
        fit(iteration).fSigma = sigmaFChoices(bestFChoice);
        % fprintf('\niteration #%d bestXYChoice = %.1f bestFChoice = %.1f\n',iteration,sigmaXYChoices(bestXYChoice),sigmaFChoices(bestFChoice))
        warpedDF = explodePadUniformXY(currentWarpedImage,sigmaXYChoices(bestXYChoice),max(4*sigmaXYChoices(bestXYChoice)+1,3),...
            sigmaXYChoices(bestXYChoice),max(3,4* ...
            sigmaXYChoices(bestXYChoice)+1),numBins);
        filter3=reshape(makeGaussianKernel(F_KERNEL_SIZE,sigmaFChoices(bestFChoice)/(256/numBins)),[1 1 F_KERNEL_SIZE]);
        
        warpedDF =convn(warpedDF, filter3, 'same');
        warpedDF = warpedDF(downsampleScale:downsampleScale:end,downsampleScale:downsampleScale:end,:);
        
        if(USE_SIC)
            warpedDFA1 = explodePadUniformXY(max(0,min(255,currentWarpedImage+1)),sigmaXYChoices(bestXYChoice),max(4*sigmaXYChoices(bestXYChoice)+1,3),...
                sigmaXYChoices(bestXYChoice),max(3,4*sigmaXYChoices(bestXYChoice)+1),numBins);
            warpedDFA1 =convn(warpedDFA1, filter3, 'same');
            warpedDFA1 = warpedDFA1(downsampleScale:downsampleScale:end,downsampleScale:downsampleScale:end,:);
            currentWarpedImageNoGain = imtransform2(img,T_temp);
            currentWarpedImageNoGain = currentWarpedImageNoGain(1:height,1:width);
            warpedDFA2 = explodePadUniformXY(max(0,min(255,currentWarpedImage+...
                currentWarpedImageNoGain*.01)),sigmaXYChoices(bestXYChoice),max(4*sigmaXYChoices(bestXYChoice)+1,3),...
                sigmaXYChoices(bestXYChoice),max(3,4*sigmaXYChoices(bestXYChoice)+1),numBins);
            warpedDFA2 =convn(warpedDFA2, filter3, 'same');
            warpedDFA2 = warpedDFA2(downsampleScale:downsampleScale:end,downsampleScale:downsampleScale:end,:);
        end
        
        wDF = shiftdim(warpedDF,2);
        tDF = shiftdim(templateDistributionFields(:,:,:,bestXYChoice,bestFChoice),2);
        
        if(direction == 1 || direction ==2)
            deltaPInverse = InverseHessians(:,:,bestXYChoice,bestFChoice)*...
                sum(steepestDescentImages(:,:,bestXYChoice,bestFChoice) .*...
                repmat(wDF(:)-tDF(:),[1 degFreedom]),1)';
            
            if(USE_SIC)
                additiveNoiseInverse = deltaPInverse(7);
                gainTempInverse = deltaPInverse(8);
            end
            deltaPTemp =reshape([deltaPInverse(1:6)],[2,3])';
            incrementalWarpInverse = [1 0 0;0 1 0;0 0 1] + [deltaPTemp [0;0;0]];
        end
        
        if(direction == 0 || direction ==2)
            [Ex2,Ey2] = gradient(warpedDF);
            steepestDescentImages2 = zeros(numBins*heightNew*widthNew,degFreedom);
            ExReshaped2 = shiftdim(Ex2,2); EyReshaped2 = shiftdim(Ey2,2);
            ExEy = [ExReshaped2(:) EyReshaped2(:)];
            for i=1:6
                steepestDescentImages2(:,i) = sum(ExEy.*dWdPReshaped(:,:,i),2);
            end
            if(USE_SIC)
                EA1Warped = (warpedDFA1 - warpedDF);
                EA2Warped = (warpedDFA2 - warpedDF)/.01;
                EA1WarpedReshaped = shiftdim(EA1Warped,2);
                EA2WarpedReshaped = shiftdim(EA2Warped,2);
                steepestDescentImages2(:,7) = EA1WarpedReshaped(:);
                steepestDescentImages2(:,8) = EA2WarpedReshaped(:);
            end
            Hessian2= steepestDescentImages2'*steepestDescentImages2;
            %Compute delta p for forward
            deltaPForward = Hessian2^(-1)*sum(steepestDescentImages2.*repmat(tDF(:)-wDF(:),[1 degFreedom]),1)';
        end
    else %use LK method
        if(direction==1 || direction ==2)
            deltaPInverse = InverseHessian*sum(steepestDescentImage .*...
                repmat(currentWarpedImage(:)-template(:),[1 degFreedom]),1)';
            
            if(USE_SIC)
                additiveNoiseInverse = deltaPInverse(7);
                gainTempInverse = deltaPInverse(8);
            end
            deltaPTemp =reshape([deltaPInverse(1:6)],[2,3])';
            incrementalWarpInverse = [1 0 0;0 1 0;0 0 1] + [deltaPTemp [0;0;0]];
        end
        
        if(direction==0 || direction ==2)
            if(additive)
                [ex,ey] = gradient(img);
                Ex2 = (imtransform2(ex,T_temp)); Ey2 = (imtransform2(ey,T_temp));
                Ex2 = Ex2(1:height,1:width); Ey2 = Ey2(1:height,1:width);
            else
                [Ex2,Ey2] = gradient(currentWarpedImage);
            end
            
            steepestDescentImages2 = zeros(numBins*heightNew*widthNew,degFreedom);
            ExEy = [Ex2(:) Ey2(:)];
            for i=1:6
                steepestDescentImages2(:,i) = sum(ExEy.*dWdPReshaped(:,:,i),2);
            end
            if(USE_SIC)
                EA1Warped = ones(size(currentWarpedImage));
                EA2Warped = (currentWarpedImage*1.01 - currentWarpedImage)/.01;
                steepestDescentImages2(:,7) = EA1Warped(:);
                steepestDescentImages2(:,8) = EA2Warped(:);
            end
            Hessian2= steepestDescentImages2'*steepestDescentImages2;
            %Compute delta p for forward
            %template has already been blurred
            deltaPForward = Hessian2^(-1)*sum(steepestDescentImages2.*repmat(template(:)-currentWarpedImage(:),[1 degFreedom]),1)';
        end
    end
    
    if(direction == 0 || direction ==2)
        deltaP2Temp =reshape([deltaPForward(1:6)],[2,3])';
        incrementalWarpForward = [1 0 0;0 1 0;0 0 1] + [deltaP2Temp [0;0;0]];
    end
    
    if(USE_SIC)
        additiveNoiseForward = deltaPForward(7);
        gainTempForward = deltaPForward(8);
        gain = gain + (gainTempForward-gainTempInverse)/2;
        additiveNoise = additiveNoise + (additiveNoiseForward-additiveNoiseInverse)/2;
    end
    
    if(direction ==2)
        incrementalWarp = real(expm((logm(incrementalWarpInverse^(-1))+ ...
            logm(incrementalWarpForward^(1)))/2));
    elseif(direction ==1)
        incrementalWarp = incrementalWarpInverse^(-1);
    else
        incrementalWarp = incrementalWarpForward;
    end
    update = incrementalWarp;
    if(additive)
        T = update-eye(3)+T;
    else
        T = update*T;
    end
    if(numBins~=1)
        terminationEPS = .003;
    else
        terminationEPS = .001;
    end
    if(iteration==NUM_ITERATIONS || sqrt(sum(sum((update-eye(3)).^2)))<terminationEPS)
        break;
    end
    iteration = iteration +1;
end %end while(true)

%Convert T to original coords
downSampledPntsOrig = [1 1 1;1 heightNew 1;widthNew 1 1;widthNew heightNew 1];
downSampledPntsNew = downSampledPntsOrig* T;
downSampledPntsNew = downSampledPntsNew./repmat(downSampledPntsNew(:,3),[1 3]);
pntsOrig = [downsampleScale* downSampledPntsOrig(:,1:2) ones(4,1)];
pntsNew = [downsampleScale* downSampledPntsNew(:,1:2) ones(4,1)];
T = pinv(pntsOrig)*pntsNew;
fit(iteration).warp_p =  T'; %save transpose to agree with Lucas-Kanade convention
fit = fit(1:iteration);

currentWarpedImage = (imtransform2(img,T)*(gain+1))+additiveNoise;
currentWarpedImage = currentWarpedImage(1:height,1:width);

currentWarpedImage(currentWarpedImage>255)=255;
currentWarpedImage(currentWarpedImage<0)=0;

if(numBins~=1)
    [xySigma,fSigma,lik] = findBestSigma(templateDistributionFields2,currentWarpedImage,...
        width,height,sigmaXYChoices,sigmaFChoices,numBins);
else
    xySigma =1; fSigma = 1; lik = 1;
end

xySigma = sigmaXYChoices(xySigma);
fSigma = sigmaFChoices(fSigma);


%use likelihood calculation to find best sigma value
function [bestXYChoice,bestFChoice,lik] = findBestSigma(templateDFs,warpedImage,width,height,sigmaXYChoices,sigmaFChoices,numBins)
binSize = 256/numBins;
distances = zeros(size(sigmaXYChoices,2),size(sigmaFChoices,2));

for m=1:size(sigmaXYChoices,2)
    for n=1:size(sigmaFChoices,2)
        [js is] = meshgrid(1:width,1:height);
        dists = max(log(templateDFs(sub2ind(size(templateDFs),is,js,floor(warpedImage(1:height,1:width)/binSize+1),...
            ones(height,width)*m,ones(height,width)*n))),log(.0001));
        dist = sum(dists(:));
        distances(m,n)=-1*dist;
    end
end

if(size(sigmaXYChoices,2)~=1)
    [vals,bestSigmaXYChoicesIndices]=min(distances);
    [val,bestFChoice]=min(vals);
    bestXYChoice = bestSigmaXYChoicesIndices(bestFChoice);
else
    [val,bestFChoice]=min(distances);
    bestXYChoice = 1;
end

lik = val;
