%EXPLODEPADUNIFORMXY computes a distribution field, smoothing it in the
%spatial dimensions
%
%   dField = explodePadUniformXY(image, sigma1, size1, sigma2, size2, bins) computes
%   a distribution field from image smoothing it in the y and x directions.
%   The DF is padded with a uniform distribution before convolution so a
%   valid convolution produces a DF that has the same number of rows and columns
%   as the original image
%   
%   Inputs:
%     1) image: Image to compute the distribution field from
%     2) sigma1: Standard deviation of the Guassian kernel used to blur in
%     the y direction
%     3) size1: Size of the Gaussian kernel used to blur in the y direction
%     4) sigma2: Standard deviation of the Guassian kernel used to blur in
%     the x direction
%     5) size2: Size of the Gaussian kernel used to blur in the x direction
%     6) bins: Number of bins to use at each pixel location when
%     constructing the distribution field
%   Outputs:
%     1) dField: The constructed distribution field

function dField = explodePadUniformXY(image, sigma1, size1, sigma2, size2, bins)
%bin size must evenly divide 256
if(mod(256,bins)~=0)
    ME = MException('BinSizeError', '256 not divisible by bins');
    throw(ME)
end

binSize = 256/bins;
dField = zeros(size(image,1),size(image,2),bins);

% generate X, Y for vectorization
[X Y] = meshgrid(1:size(image, 2), 1:size(image, 1));

% bin intensity values
I = floor(double(image)/binSize)+1;

% build up histogram by indexing vectorization on y, x, intensity
dField(sub2ind(size(dField), Y(:), X(:), I(:))) = 1;

dFieldTemp = zeros(size(image,1)+size1-1,size(image,2)+size2-1,bins);

dFieldTemp((size1-1)/2+1:(size1-1)/2+size(image,1),...
	   (size2-1)/2+1:(size2-1)/2+size(image,2),:) = dField;
   
%Pad with uniform distribution
dFieldTemp(1:(size1-1)/2,:,:)=1/bins;
dFieldTemp((size1-1)/2+size(image,1)+1:end,:,:) = 1/bins;

dFieldTemp(:,1:(size2-1)/2,:)=1/bins;
dFieldTemp(:,(size2-1)/2+size(image,2)+1:end,:) = 1/bins;


filter1=reshape(makeGaussianKernel(size1,sigma1),[size1 1]);
filter2=reshape(makeGaussianKernel(size2,sigma2),[1 size2]);

dField =  convn(convn(dFieldTemp,filter1,'valid'),filter2,'valid');



